@@ -1,5 +1,9 @@
 # -*- coding: utf-8 -*-
-"""Multipath Fading Channel prebuilt templates."""
+"""
+===================================
+Multipath Fading Standard Templates
+===================================
+"""
 
 from __future__ import annotations
 import numpy as np
@@ -10,12 +14,12 @@ from ruamel.yaml import SafeConstructor, SafeRepresenter, MappingNode, ScalarNod
 from hermespy.channel import MultipathFadingChannel
 
 __author__ = "Tobias Kronauer"
-__copyright__ = "Copyright 2021, Barkhausen Institut gGmbH"
+__copyright__ = "Copyright 2022, Barkhausen Institut gGmbH"
 __credits__ = ["Tobias Kronauer", "Jan Adler"]
 __license__ = "AGPLv3"
-__version__ = "0.2.4"
-__maintainer__ = "Tobias Kronauer"
-__email__ = "tobias.kronaue@barkhauseninstitut.org"
+__version__ = "0.2.7"
+__maintainer__ = "Jan Adler"
+__email__ = "jan.adler@barkhauseninstitut.org"
 __status__ = "Prototype"
 
 
@@ -142,8 +146,6 @@ class MultipathFadingCost256(MultipathFadingChannel):
             'los_angle': node.los_angle,
             'doppler_frequency': node.doppler_frequency,
             'los_doppler_frequency': node.los_doppler_frequency,
-            'transmit_precoding': node.transmit_precoding,
-            'receive_postcoding': node.receive_postcoding,
             'sync_offset_low': node.sync_offset_low,
             'sync_offset_high': node.sync_offset_high
         }
@@ -151,10 +153,7 @@ class MultipathFadingCost256(MultipathFadingChannel):
         if node.model_type is MultipathFadingCost256.TYPE.HILLY:
             state.pop('los_angle')
 
-        transmitter_index, receiver_index = node.indices
-
-        yaml = representer.represent_mapping(u'{.yaml_tag} {} {}'.format(cls, transmitter_index, receiver_index), state)
-        return yaml
+        return representer.represent_mapping(cls.yaml_tag, state)
 
     @classmethod
     def from_yaml(cls: Type[MultipathFadingCost256], constructor: SafeConstructor, node: MappingNode) -> \
@@ -356,8 +355,6 @@ class MultipathFading5GTDL(MultipathFadingChannel):
             'los_angle': node.los_angle,
             'doppler_frequency': node.doppler_frequency,
             'los_doppler_frequency': node.los_doppler_frequency,
-            'transmit_precoding': node.transmit_precoding,
-            'receive_postcoding': node.receive_postcoding,
             'sync_offset_low': node.sync_offset_low,
             'sync_offset_high': node.sync_offset_high
         }
@@ -365,10 +362,7 @@ class MultipathFading5GTDL(MultipathFadingChannel):
         if node.model_type is MultipathFading5GTDL.TYPE.C or MultipathFading5GTDL.TYPE.E:
             state.pop('los_doppler_frequency')
 
-        transmitter_index, receiver_index = node.indices
-
-        yaml = representer.represent_mapping(u'{.yaml_tag} {} {}'.format(cls, transmitter_index, receiver_index), state)
-        return yaml
+        return representer.represent_mapping(cls.yaml_tag, state)
 
     @classmethod
     def from_yaml(cls: Type[MultipathFading5GTDL], constructor: SafeConstructor, node: MappingNode) -> \
@@ -490,14 +484,11 @@ class MultipathFadingExponential(MultipathFadingChannel):
             'los_angle': node.los_angle,
             'doppler_frequency': node.doppler_frequency,
             'los_doppler_frequency': node.los_doppler_frequency,
-            'transmit_precoding': node.transmit_precoding,
-            'receive_postcoding': node.receive_postcoding,
             'sync_offset_low': node.sync_offset_low,
             'sync_offset_high': node.sync_offset_high
         }
 
-        transmitter_index, receiver_index = node.indices
-        return representer.represent_mapping(u'{.yaml_tag} {} {}'.format(cls, transmitter_index, receiver_index), state)
+        return representer.represent_mapping(cls.yaml_tag, state)
 
     @classmethod
     def from_yaml(cls: Type[MultipathFadingExponential], constructor: SafeConstructor, node: MappingNode) -> \
