@@ -1,24 +1,24 @@
-from .modem import TransmissionMode, Modem
-from .transmitter import Transmitter
-from .receiver import Receiver
-from .rf_chain import RfChain
+from .bits_source import RandomBitsSource, StreamBitsSource
+from .modem import Modem
 from .waveform_generator import WaveformGenerator, Synchronization
 from .waveform_generator_chirp_fsk import WaveformGeneratorChirpFsk, ChirpFskSynchronization,\
     ChirpFskCorrelationSynchronization
 from .waveform_generator_psk_qam import WaveformGeneratorPskQam
 from .waveform_generator_ofdm import WaveformGeneratorOfdm, FrameGuardSection, FrameSymbolSection, FrameResource
 from .tools.shaping_filter import ShapingFilter
+from .evaluators import BitErrorEvaluator, BlockErrorEvaluator, FrameErrorEvaluator, ThroughputEvaluator
 
 __author__ = "Jan Adler"
-__copyright__ = "Copyright 2021, Barkhausen Institut gGmbH"
+__copyright__ = "Copyright 2022, Barkhausen Institut gGmbH"
 __credits__ = ["Jan Adler"]
 __license__ = "AGPLv3"
-__version__ = "0.2.4"
+__version__ = "0.2.7"
 __maintainer__ = "Jan Adler"
 __email__ = "jan.adler@barkhauseninstitut.org"
 __status__ = "Prototype"
 
 
-__all__ = ['Modem', 'Transmitter', 'Receiver', 'RfChain', 'WaveformGenerator', 'WaveformGeneratorChirpFsk',
+__all__ = ['RandomBitsSource', 'StreamBitsSource', 'Modem', 'WaveformGenerator', 'WaveformGeneratorChirpFsk',
            'WaveformGeneratorPskQam', 'WaveformGeneratorOfdm', 'ShapingFilter', 'FrameGuardSection',
-           'FrameSymbolSection', 'FrameResource', 'Synchronization']
+           'FrameSymbolSection', 'FrameResource', 'Synchronization', 'BitErrorEvaluator', 'BlockErrorEvaluator',
+           'FrameErrorEvaluator', 'ThroughputEvaluator']
