@@ -1,37 +1,48 @@
 # -*- coding: utf-8 -*-
-"""Prototype for Waveform Generation Modeling."""
+"""
+===========================
+Communication Waveform Base
+===========================
+"""
 
 from __future__ import annotations
 from abc import ABC, abstractmethod
-from math import floor, sqrt
-from typing import Tuple, TYPE_CHECKING, Optional, Type, List
+from math import floor
+from typing import Generic, Tuple, TYPE_CHECKING, Optional, Type, TypeVar, List
 
 import numpy as np
 from ruamel.yaml import SafeConstructor, SafeRepresenter, Node
 
-from hermespy.channel import ChannelStateInformation
-from hermespy.signal import Signal
+from hermespy.core.channel_state_information import ChannelStateInformation
+from hermespy.core.signal_model import Signal
+from .symbols import Symbols
 
 if TYPE_CHECKING:
     from hermespy.modem import Modem
 
-__author__ = "Tobias Kronauer"
+__author__ = "Andre Noll Barreto"
 __copyright__ = "Copyright 2021, Barkhausen Institut gGmbH"
-__credits__ = ["Tobias Kronauer", "Jan Adler"]
+__credits__ = ["Andre Noll Barreto", "Tobias Kronauer", "Jan Adler"]
 __license__ = "AGPLv3"
-__version__ = "0.2.4"
-__maintainer__ = "Tobias Kronauer"
-__email__ = "tobias.kronauer@barkhauseninstitut.org"
+__version__ = "0.2.7"
+__maintainer__ = "Jan Adler"
+__email__ = "jan.adler@barkhauseninstitut.org"
 __status__ = "Prototype"
 
 
-class Synchronization(ABC):
-    """Abstract base class for synchronization routines of waveform generators."""
+WaveformType = TypeVar('WaveformType', bound='WaveformGenerator')
 
-    __waveform_generator: Optional[WaveformGenerator]       # Waveform generator this routine is attached to
+
+class Synchronization(Generic[WaveformType], ABC):
+    """Abstract base class for synchronization routines of waveform generators.
+
+    Refer to :footcite:t:`2016:nasir` for an overview of the current state of the art.
+    """
+
+    __waveform_generator: Optional[WaveformType]       # Waveform generator this routine is attached to
 
     def __init__(self,
-                 waveform_generator: Optional[WaveformGenerator] = None) -> None:
+                 waveform_generator: Optional[WaveformType] = None) -> None:
         """
         Args:
             waveform_generator (WaveformGenerator, optional):
@@ -41,18 +52,18 @@ class Synchronization(ABC):
         self.__waveform_generator = waveform_generator
 
     @property
-    def waveform_generator(self) -> Optional[WaveformGenerator]:
+    def waveform_generator(self) -> Optional[WaveformType]:
         """Waveform generator this synchronization routine is attached to.
 
         Returns:
-            Optional[WaveformGenerator]:
+            Optional[WaveformType]:
                 Handle to the waveform generator. None if the synchronization routine is floating.
         """
 
         return self.__waveform_generator
 
     @waveform_generator.setter
-    def waveform_generator(self, value: Optional[WaveformGenerator]) -> None:
+    def waveform_generator(self, value: Optional[WaveformType]) -> None:
         """Set waveform generator this synchronization routine is attached to."""
 
         if self.__waveform_generator is not None:
@@ -114,6 +125,119 @@ class Synchronization(ABC):
         return synchronized_frames
 
 
+class ChannelEstimation(Generic[WaveformType], ABC):
+    """Abstract base class for channel estimation routines of waveform generators."""
+
+    def __init__(self,
+                 waveform_generator: Optional[WaveformType] = None) -> None:
+        """
+        Args:
+            waveform_generator (WaveformGenerator, optional):
+                The waveform generator this estimation routine is attached to.
+        """
+
+        self.__waveform_generator = waveform_generator
+
+    @property
+    def waveform_generator(self) -> Optional[WaveformType]:
+        """Waveform generator this synchronization routine is attached to.
+
+        Returns:
+            Optional[WaveformType]:
+                Handle to the waveform generator. None if the synchronization routine is floating.
+        """
+
+        return self.__waveform_generator
+
+    @waveform_generator.setter
+    def waveform_generator(self, value: Optional[WaveformType]) -> None:
+        """Set waveform generator this synchronization routine is attached to."""
+
+        if self.__waveform_generator is not None:
+            raise RuntimeError("Error trying to re-attach already attached synchronization routine.")
+
+        self.__waveform_generator = value
+
+    def estimate_channel(self,
+                         signal: Signal,
+                         csi: Optional[ChannelStateInformation] = None) -> ChannelStateInformation:
+        """Estimate the wireless channel of a received communication frame.
+
+        Args:
+
+            signal (Signal):
+                Signal model of the communication frame waveform.
+
+            csi (ChannelStateInformation, optional):
+                Ideal channel state information.
+                May be required for some routines.
+
+        Raises:
+            ValueError:
+                If `csi` is required but not provided.
+        """
+
+        if csi is None:
+            raise ValueError("Ideal channel estimation requires prior channel state information")
+
+        return csi
+    
+
+class ChannelEqualization(Generic[WaveformType], ABC):
+    """Abstract base class for channel equalization routines of waveform generators."""
+
+    def __init__(self,
+                 waveform_generator: Optional[WaveformType] = None) -> None:
+        """
+        Args:
+            waveform_generator (WaveformGenerator, optional):
+                The waveform generator this equalization routine is attached to.
+        """
+
+        self.__waveform_generator = waveform_generator
+
+    @property
+    def waveform_generator(self) -> Optional[WaveformType]:
+        """Waveform generator this equalization routine is attached to.
+
+        Returns:
+            Optional[WaveformType]:
+                Handle to the waveform generator. None if the equalization routine is floating.
+        """
+
+        return self.__waveform_generator
+
+    @waveform_generator.setter
+    def waveform_generator(self, value: Optional[WaveformType]) -> None:
+        """Set waveform generator this equalization routine is attached to."""
+
+        if self.__waveform_generator is not None:
+            raise RuntimeError("Error trying to re-attach already attached equalization routine.")
+
+        self.__waveform_generator = value
+
+    def equalize_channel(self,
+                         signal: Signal,
+                         csi: ChannelStateInformation) -> Signal:
+        """Equalize the wireless channel of a received communication frame.
+
+        Args:
+
+            signal (Signal):
+                Signal model of the communication frame waveform.
+
+            csi (ChannelStateInformation):
+                Channel state estimation
+
+        Returns:
+            Signal:
+                The equalized signal model.
+        """
+        
+        # The default routine performs no equalization
+        return signal
+
+
 class WaveformGenerator(ABC):
     """Implements an abstract waveform generator.
 
@@ -123,15 +247,15 @@ class WaveformGenerator(ABC):
     yaml_tag: str = "Waveform"
     """YAML serialization tag."""
 
-    synchronization: Synchronization
-    """Synchronization routine."""
-
     symbol_type: np.dtype = complex
     """Symbol type."""
 
-    __modem: Optional[Modem]            # Modem this waveform generator is attached to
-    __oversampling_factor: int          # Oversampling factor
-    __modulation_order: int             # Cardinality of the set of communication symbols
+    __modem: Optional[Modem]                        # Modem this waveform generator is attached to
+    __synchronization: Synchronization              # Synchronization routine
+    __channel_estimation: ChannelEstimation         # Channel estimation routine
+    __channel_equalization: ChannelEqualization     # Channel equalization routine
+    __oversampling_factor: int                      # Oversampling factor
+    __modulation_order: int                         # Cardinality of the set of communication symbols
 
     def __init__(self,
                  modem: Optional[Modem] = None,
@@ -157,6 +281,8 @@ class WaveformGenerator(ABC):
         self.oversampling_factor = oversampling_factor
         self.modulation_order = modulation_order
         self.synchronization = Synchronization(self)
+        self.channel_estimation = ChannelEstimation(self)
+        self.channel_equalization = ChannelEqualization(self)
 
         if modem is not None:
             self.modem = modem
@@ -246,7 +372,7 @@ class WaveformGenerator(ABC):
             int: Number of bits per symbol
         """
 
-        return int(sqrt(self.__modulation_order))
+        return int(np.log2(self.__modulation_order))
 
     @property
     @abstractmethod
@@ -312,7 +438,7 @@ class WaveformGenerator(ABC):
         ...
 
     @abstractmethod
-    def map(self, data_bits: np.ndarray) -> np.ndarray:
+    def map(self, data_bits: np.ndarray) -> Symbols:
         """Map a stream of bits to data symbols.
 
         Args:
@@ -320,19 +446,17 @@ class WaveformGenerator(ABC):
                 Vector containing a sequence of L hard data bits to be mapped onto data symbols.
 
         Returns:
-            np.ndarray:
-                Vector containing the resulting sequence of K data symbols.
-                In general, K is less or equal to L.
+            Symbols: Mapped data symbols.
         """
         ...
 
     @abstractmethod
-    def unmap(self, data_symbols: np.ndarray) -> np.ndarray:
+    def unmap(self, symbols: Symbols) -> np.ndarray:
         """Map a stream of data symbols to data bits.
 
         Args:
-            data_symbols (np.ndarray):
-                Vector containing a sequence of K data symbols to be mapped onto bit sequences.
+            symbols (Symbols):
+                Sequence of K data symbols to be mapped onto bit sequences.
 
         Returns:
             np.ndarray:
@@ -361,7 +485,7 @@ class WaveformGenerator(ABC):
     def demodulate(self,
                    signal: np.ndarray,
                    channel_state: ChannelStateInformation,
-                   noise_variance: float) -> Tuple[np.ndarray, ChannelStateInformation, np.ndarray]:
+                   noise_variance: float) -> Tuple[Symbols, ChannelStateInformation, np.ndarray]:
         """Demodulate a base-band signal stream to data symbols.
 
         Args:
@@ -429,6 +553,60 @@ class WaveformGenerator(ABC):
             handle.waveform_generator = self
 
         self.__modem = handle
+        
+    @property
+    def synchronization(self) -> Synchronization:
+        """Synchronization routine.
+
+        Returns:
+            Synchronization: Handle to the synchronization routine.
+        """
+
+        return self.__synchronization
+
+    @synchronization.setter
+    def synchronization(self, value: Synchronization) -> None:
+
+        self.__synchronization = value
+
+        if value.waveform_generator is not self:
+            value.waveform_generator = self
+
+    @property
+    def channel_estimation(self) -> ChannelEstimation:
+        """Channel estimation routine.
+
+        Returns:
+            ChannelEstimation: Handle to the synchronization routine.
+        """
+
+        return self.__channel_estimation
+
+    @channel_estimation.setter
+    def channel_estimation(self, value: ChannelEstimation) -> None:
+
+        self.__channel_estimation = value
+
+        if value.waveform_generator is not self:
+            value.waveform_generator = self
+            
+    @property
+    def channel_equalization(self) -> ChannelEqualization:
+        """Channel estimation routine.
+
+        Returns:
+            ChannelEqualization: Handle to the equalization routine.
+        """
+
+        return self.__channel_equalization
+
+    @channel_equalization.setter
+    def channel_equalization(self, value: ChannelEqualization) -> None:
+
+        self.__channel_equalization = value
+
+        if value.waveform_generator is not self:
+            value.waveform_generator = self
 
     @property
     @abstractmethod
@@ -482,3 +660,17 @@ class WaveformGenerator(ABC):
 
         state = constructor.construct_mapping(node)
         return cls(**state)
+
+
+class PilotWaveformGenerator(WaveformGenerator, ABC):
+    """Abstract base class of communication waveform generators generating a pilot sequence."""
+
+    @property
+    @abstractmethod
+    def pilot(self) -> Signal:
+        """Model of the pilot sequence within this communication waveform.
+
+        Returns:
+            Signal: The pilot sequence.
+        """
+        ...
