@@ -1,35 +1,41 @@
 # -*- coding: utf-8 -*-
-"""Chirp Frequency Shift Keying Waveform Generator."""
+"""
+============================
+Chirp Frequency Shift Keying
+============================
+"""
 
 from __future__ import annotations
-from typing import Any, List, Optional, Tuple, Type
+from typing import Optional, Tuple, Type
 from math import ceil
 from functools import lru_cache
 
 import numpy as np
 from ruamel.yaml import SafeConstructor, SafeRepresenter, Node
 from scipy import integrate
-from scipy.signal import correlate
 
+from hermespy.core.factory import Serializable
 from hermespy.channel import ChannelStateInformation
-from hermespy.modem.waveform_generator import WaveformGenerator, Synchronization
-from hermespy.signal import Signal
+from hermespy.modem.waveform_generator import PilotWaveformGenerator, WaveformGenerator, Synchronization
+from hermespy.core.signal_model import Signal
+from .symbols import Symbols
+from .waveform_correlation_synchronization import CorrelationSynchronization
 
-__author__ = "Tobias Kronauer"
+__author__ = "Andre Noll Barreto"
 __copyright__ = "Copyright 2021, Barkhausen Institut gGmbH"
-__credits__ = ["Tobias Kronauer", "Jan Adler"]
+__credits__ = ["Andre Noll Barreto", "Tobias Kronauer", "Jan Adler"]
 __license__ = "AGPLv3"
-__version__ = "0.2.4"
-__maintainer__ = "Tobias Kronauer"
-__email__ = "tobias.kronauer@barkhauseninstitut.org"
+__version__ = "0.2.7"
+__maintainer__ = "Jan Adler"
+__email__ = "jan.adler@barkhauseninstitut.org"
 __status__ = "Prototype"
 
 
-class WaveformGeneratorChirpFsk(WaveformGenerator):
+class WaveformGeneratorChirpFsk(PilotWaveformGenerator, Serializable):
     """ Implements a chirp FSK waveform generator."""
 
     # YAML tag
-    yaml_tag = WaveformGenerator.yaml_tag + "ChirpFsk"
+    yaml_tag = u'ChirpFsk'
 
     # Modulation parameters
     symbol_type: np.dtype = int
@@ -69,10 +75,10 @@ class WaveformGeneratorChirpFsk(WaveformGenerator):
         WaveformGenerator.__init__(self, **kwargs)
 
         # Default parameters
-        self.synchronization = ChirpFskSynchronization(self)
-        self.__chirp_duration = 512e-6
-        self.__chirp_bandwidth = 500e3
-        self.__freq_difference = 1953.125
+        self.synchronization = ChirpFskSynchronization()
+        self.__chirp_duration = 1e-6
+        self.__chirp_bandwidth = 10e6
+        self.__freq_difference = self.__chirp_bandwidth / self.modulation_order
         self.__num_pilot_chirps = 0
         self.__num_data_chirps = 20
         self.__guard_interval = 0.0
@@ -432,12 +438,12 @@ class WaveformGeneratorChirpFsk(WaveformGenerator):
         bit_energy = symbol_energy / self.bits_per_symbol
         return bit_energy
 
-    def map(self, data_bits: np.ndarray) -> np.ndarray:
+    def map(self, data_bits: np.ndarray) -> Symbols:
 
         offset = self._calculate_frequency_offsets(data_bits)
-        return offset
+        return Symbols(offset)
 
-    def modulate(self, data_symbols: np.ndarray) -> Signal:
+    def modulate(self, data_symbols: Symbols) -> Signal:
 
         prototypes, _ = self._prototypes()
         samples = np.empty(self.samples_in_frame, dtype=complex)
@@ -446,21 +452,23 @@ class WaveformGeneratorChirpFsk(WaveformGenerator):
         samples_in_chirp = self.samples_in_chirp
 
         # Add pilot samples
-        samples[:self.pilot_samples.shape[0]] = self.pilot_samples
-        sample_idx += self.pilot_samples.shape[0]
+        pilot_samples = self.pilot.samples.flatten()
+        num_pilot_samples = len(pilot_samples)
+        samples[:num_pilot_samples] = pilot_samples
+        sample_idx += num_pilot_samples
 
         # Modulate data symbols
-        for symbol in data_symbols:
+        for symbol in data_symbols.raw[0, :]:
 
             samples[sample_idx:sample_idx+samples_in_chirp] = prototypes[symbol, :]
             sample_idx += samples_in_chirp
 
-        return Signal(samples, self.sampling_rate, carrier_frequency=self.modem.carrier_frequency)
+        return Signal(samples, self.sampling_rate)
 
     def demodulate(self,
                    baseband_signal: np.ndarray,
                    channel_state: ChannelStateInformation,
-                   noise_variance: float) -> Tuple[np.ndarray, ChannelStateInformation, np.ndarray]:
+                   noise_variance: float) -> Tuple[Symbols, ChannelStateInformation, np.ndarray]:
 
         # Assess number of frames contained within this signal
         samples_in_chirp = self.samples_in_chirp
@@ -479,14 +487,14 @@ class WaveformGeneratorChirpFsk(WaveformGenerator):
                                                       channel_state.num_receive_streams)
         noises = np.repeat(noise_variance, self.num_data_chirps)
 
-        return symbols, channel_state, noises
+        return Symbols(symbols), channel_state, noises
 
-    def unmap(self, data_symbols: np.ndarray) -> np.ndarray:
+    def unmap(self, data_symbols: Symbols) -> np.ndarray:
 
         bits_per_symbol = self.bits_per_symbol
-        bits = np.empty(data_symbols.shape[0] * self.bits_per_symbol)
+        bits = np.empty(data_symbols.num_symbols * self.bits_per_symbol)
 
-        for s, symbol in enumerate(data_symbols):
+        for s, symbol in enumerate(data_symbols.raw[0, :]):
 
             symbol_bits = [int(x) for x in list(np.binary_repr(int(symbol.real), width=bits_per_symbol))]
             bits[s*bits_per_symbol:(s+1)*bits_per_symbol] = symbol_bits
@@ -563,9 +571,8 @@ class WaveformGeneratorChirpFsk(WaveformGenerator):
         # Sampling rate scales with the chirp bandwidth
         return self.oversampling_factor * self.__chirp_bandwidth
 
-   # @cached_property
     @property
-    def pilot_samples(self) -> np.ndarray:
+    def pilot(self) -> Signal:
         """Samples of the frame's pilot section.
 
         Returns:
@@ -574,146 +581,35 @@ class WaveformGeneratorChirpFsk(WaveformGenerator):
 
         # Generate single pilot chirp prototype
         prototypes, _ = self._prototypes()
-        pilot_chirp = prototypes[0]
 
-        samples = np.tile(pilot_chirp, self.__num_pilot_chirps)
-        return samples
+        pilot_samples = np.empty(self.samples_in_chirp * self.num_pilot_chirps, dtype=complex)
+        for pilot_idx in range(self.num_pilot_chirps):
+
+            pilot_samples[pilot_idx*self.samples_in_chirp:(1+pilot_idx)*self.samples_in_chirp] = \
+                prototypes[pilot_idx % len(prototypes)]
+
+        return Signal(pilot_samples, self.sampling_rate)
 
     def _clear_cache(self) -> None:
         """Clear cached properties because a parameter has changed."""
-
-        #if self._pilot_samples:
-        #    del self._pilot_samples
+        pass
 
 
-class ChirpFskSynchronization(Synchronization):
+class ChirpFskSynchronization(Synchronization[WaveformGeneratorChirpFsk]):
     """Synchronization for chirp-based frequency shift keying communication waveforms."""
 
-    __waveform_generator: Optional[WaveformGeneratorChirpFsk]
-
     def __init__(self,
-                 waveform_generator: Optional[WaveformGenerator] = None,
-                 *args: Any) -> None:
+                 waveform_generator: Optional[WaveformGeneratorChirpFsk] = None) -> None:
         """
         Args:
 
             waveform_generator (WaveformGenerator, optional):
                 The waveform generator this synchronization routine is attached to.
-
-            *args:
-                Synchronization base class initialization parameters.
         """
 
-        self.__waveform_generator = waveform_generator
-        Synchronization.__init__(self, *args)
-
-    @property
-    def waveform_generator(self) -> Optional[WaveformGeneratorChirpFsk]:
-
-        return self.__waveform_generator
-
-    @waveform_generator.setter
-    def waveform_generator(self, value: Optional[WaveformGeneratorChirpFsk]) -> None:
-
-        if self.__waveform_generator is not None:
-            raise RuntimeError("Error trying to re-attach already attached synchronization routine.")
-
-        self.__waveform_generator = value
+        Synchronization.__init__(self, waveform_generator)
 
 
-class ChirpFskCorrelationSynchronization(ChirpFskSynchronization):
-    """Correlation-based clock synchronization for chirp frequency shift keying waveforms."""
-
-    __threshold: float      # Correlation threshold at which a pilot signal is detected
-    __guard_ratio: float    # Guard ratio of frame duration
-
-    def __init__(self,
-                 threshold: float = 0.9,
-                 guard_ratio: float = 0.8,
-                 *args: Any,
-                 **kwargs: Any) -> None:
-        """
-        Args:
-            
-            threshold (float, optional):
-                Correlation threshold at which a pilot signal is detected.
-                
-            guard_ratio (float, optional):
-                Guard ratio of frame duration.
-            
-            *args:
-                Synchronization base class initialization parameters.
-        """
-
-        self.threshold = threshold
-        self.guard_ratio = guard_ratio
-
-        ChirpFskSynchronization.__init__(self, *args, **kwargs)
-
-    @property
-    def threshold(self) -> float:
-        """Correlation threshold at which a pilot signal is detected.
-        
-        Returns:
-            float: Threshold between zero and one.
-            
-        Raises:
-            ValueError: If threshold is smaller than zero or greater than one.
-        """
-        
-        return self.__threshold
-    
-    @threshold.setter
-    def threshold(self, value: float):
-        """Set correlation threshold at which a pilot signal is detected."""
-        
-        if value < 0. or value > 1.:
-            raise ValueError("Synchronization threshold must be between zero and one.")
-        
-        self.__threshold = value
-
-    @property
-    def guard_ratio(self) -> float:
-        """Correlation guard ratio at which a pilot signal is detected.
-
-        After the detection of a pilot section, `guard_ratio` prevents the detection of another pilot in
-        the following samples for a span relative to the configured frame duration.
-
-        Returns:
-            float: Guard Ratio between zero and one.
-
-        Raises:
-            ValueError: If guard ratio is smaller than zero or greater than one.
-        """
-
-        return self.__guard_ratio
-
-    @guard_ratio.setter
-    def guard_ratio(self, value: float):
-        """Set correlation guard ratio at which a pilot signal is detected."""
-
-        if value < 0. or value > 1.:
-            raise ValueError("Synchronization guard ratio must be between zero and one.")
-
-        self.__guard_ratio = value
-
-    def synchronize(self,
-                    signal: np.ndarray,
-                    channel_state: ChannelStateInformation) -> List[Tuple[np.ndarray, ChannelStateInformation]]:
-
-        # Query the pilot signal from the waveform generator
-        pilot = self.waveform_generator.pilot_samples
-
-        # If no pilot signal is generated, no correlation can be done
-        if pilot.shape[0] < 1:
-            raise RuntimeError("Waveform generator does not generate a pilot signal, correlation synchronization failed")
-
-        correlation = correlate(signal, pilot, mode='same', method='fft')
-        correlation /= (np.linalg.norm(self.waveform_generator.pilot_samples) ** 2)  # Normalize correlation
-
-        pilot_indices = np.argwhere(np.greater_equal(correlation, self.__threshold * correlation.max())).flatten()
-        pilot_indices -= int(.5 * pilot.shape[0])
-        frame_length = self.waveform_generator.samples_in_frame
-        frames = [(signal[idx:idx+frame_length], channel_state[idx:idx+frame_length]) for idx in pilot_indices]
-
-        return frames
+class ChirpFskCorrelationSynchronization(CorrelationSynchronization[WaveformGeneratorChirpFsk]):
+    """Correlation-based clock-synchronization for Chirp-FSK waveforms."""
+    ...
