@@ -1,5 +1,9 @@
 # -*- coding: utf-8 -*-
-"""HermesPy Orthogonal Frequency Division Multiplexing Waveform Generation."""
+"""
+==========================================
+Orthogonal Frequency Division Multiplexing
+==========================================
+"""
 
 from __future__ import annotations
 from typing import List, Tuple, Optional, Type, Union, Any
@@ -11,16 +15,17 @@ from ruamel.yaml import SafeConstructor, SafeRepresenter, MappingNode, ScalarNod
 from scipy.fft import fft, ifft
 from scipy.interpolate import griddata
 
-from hermespy.channel import ChannelStateInformation, ChannelStateDimension
-from hermespy.modem import WaveformGenerator
-from hermespy.modem.tools import PskQamMapping
-from hermespy.signal import Signal
+from ..core.factory import Serializable
+from ..core.channel_state_information import ChannelStateInformation, ChannelStateDimension
+from ..core.signal_model import Signal
+from .modem import Symbols, WaveformGenerator
+from .tools import PskQamMapping
 
 __author__ = "André Noll Barreto"
 __copyright__ = "Copyright 2021, Barkhausen Institut gGmbH"
-__credits__ = ["André Barreto", "Jan Adler"]
+__credits__ = ["André Noll Barreto", "Tobias Kronauer", "Jan Adler"]
 __license__ = "AGPLv3"
-__version__ = "0.2.4"
+__version__ = "0.2.7"
 __maintainer__ = "Jan Adler"
 __email__ = "jan.adler@barkhauseninstitut.org"
 __status__ = "Prototype"
@@ -323,7 +328,7 @@ class FrameSection:
         ...
 
 
-class FrameSymbolSection(FrameSection):
+class FrameSymbolSection(FrameSection, Serializable):
 
     yaml_tag: str = u'Symbol'
     pattern: List[int]
@@ -379,7 +384,7 @@ class FrameSymbolSection(FrameSection):
         mask = self.resource_mask
 
         # Fill up the time-frequency grid exploiting the mask
-        grid = np.zeros((self.frame.num_subcarriers, self.num_words), complex)
+        grid = np.zeros((self.frame.num_subcarriers, self.num_words), dtype=complex)
 
         # Reference fields all currently carry the complex symbol 1+j0
         # ToDo: Implement reference symbol configurations
@@ -482,8 +487,18 @@ class FrameSymbolSection(FrameSection):
 
         return cls(**constructor.construct_mapping(node))
 
+    @classmethod
+    def to_yaml(cls: Type[FrameSymbolSection], representer: SafeRepresenter, node: FrameSymbolSection) -> MappingNode:
 
-class FrameGuardSection(FrameSection):
+        state = {
+            'num_repetitions': node.num_repetitions,
+            'pattern': node.pattern,
+        }
+
+        return representer.represent_mapping(node.yaml_tag, state)
+
+
+class FrameGuardSection(FrameSection, Serializable):
 
     yaml_tag: str = u'Guard'
     __duration: float
@@ -548,8 +563,18 @@ class FrameGuardSection(FrameSection):
 
         return cls(**constructor.construct_mapping(node))
 
+    @classmethod
+    def to_yaml(cls: Type[FrameGuardSection], representer: SafeRepresenter, node: FrameGuardSection) -> MappingNode:
 
-class WaveformGeneratorOfdm(WaveformGenerator):
+        state = {
+            'num_repetitions': node.num_repetitions,
+            'duration': node.duration,
+        }
+
+        return representer.represent_mapping(cls.yaml_tag, state)
+
+
+class WaveformGeneratorOfdm(WaveformGenerator, Serializable):
     """Generic Orthogonal-Frequency-Division-Multiplexing with a flexible frame configuration.
 
     The following features are supported:
@@ -585,7 +610,7 @@ class WaveformGeneratorOfdm(WaveformGenerator):
             Time-domain frame configuration.
     """
 
-    yaml_tag: str = WaveformGenerator.yaml_tag + u'OFDM'
+    yaml_tag: str = u'OFDM'
 
     __channel_estimation_algorithm: ChannelEstimation
     __subcarrier_spacing: float
@@ -769,19 +794,22 @@ class WaveformGeneratorOfdm(WaveformGenerator):
 
         return num
 
-    def map(self, data_bits: np.ndarray) -> np.ndarray:
-        return self._mapping.get_symbols(data_bits)
+    def map(self, data_bits: np.ndarray) -> Symbols:
+        return Symbols(self._mapping.get_symbols(data_bits))
 
-    def unmap(self, data_symbols: np.ndarray) -> np.ndarray:
+    def unmap(self, data_symbols: Symbols) -> np.ndarray:
 
-        detected_bits = self._mapping.detect_bits(data_symbols).astype(int)
+        detected_bits = self._mapping.detect_bits(data_symbols.raw).astype(int)
         return detected_bits
 
-    def modulate(self, data_symbols: np.ndarray) -> Signal:
+    def modulate(self, data_symbols: Symbols) -> Signal:
 
         # The number of samples in time domain the frame should contain, given the current sample frequency
         output_signal = np.empty(0, dtype=complex)
 
+        # Convert symbols
+        data_symbols = data_symbols.raw.flatten()
+
         sent_data_symbols = 0
         for section in self.structure:
 
@@ -796,13 +824,13 @@ class WaveformGeneratorOfdm(WaveformGenerator):
             section_signal = section.modulate(section_data_symbols)
             output_signal = np.append(output_signal, section_signal)
 
-        signal_model = Signal(output_signal, self.sampling_rate, carrier_frequency=self.modem.carrier_frequency)
+        signal_model = Signal(output_signal, self.sampling_rate)
         return signal_model
 
     def demodulate(self,
                    signal: np.ndarray,
                    channel_state: ChannelStateInformation,
-                   noise_variance: float = 0.0) -> Tuple[np.ndarray, ChannelStateInformation, np.ndarray]:
+                   noise_variance: float = 0.0) -> Tuple[Symbols, ChannelStateInformation, np.ndarray]:
 
         # Recover OFDM grid
         symbol_grid = np.empty((self.num_subcarriers, self.words_per_frame), dtype=complex)
@@ -848,7 +876,7 @@ class WaveformGeneratorOfdm(WaveformGenerator):
         data_symbols = symbol_grid.T[data_mask.T]
         noise_variances = np.repeat(noise_variance, self.symbols_per_frame)
 
-        return data_symbols, channel_state_estimation, noise_variances
+        return Symbols(data_symbols), channel_state_estimation, noise_variances
 
     @property
     def bandwidth(self) -> float:
